"""Module containing python functions, which solve 2vN approach integral equations."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
from scipy import pi
import itertools

from ..mytypes import complexnp
from ..mytypes import doublenp
from ..mytypes import intnp

from ..specfunc import kernel_fredriksen
from ..specfunc import hilbert_fredriksen
from ..aprclass import Approach2vN

def func_2vN(Ek, Ek_grid, l, eta, hfk):
    """
    Linearly interpolate the value of hfk on Ek_grid at point Ek.

    Parameters
    ----------
    Ek : float
        Energy value (not necessarily a grid point).
    Ek_grid : array
        Energy grid.
    l : int
        Lead label.
    eta : int
        Integer describing (+/-1) if infinitesimal eta is positive or negative.
    hfk : array
        Array containing Hilbert transform of Fermi function (or 1-Fermi).

    Returns
    -------
    float
        Interpolated value of hfk at Ek.
    """
    if Ek<Ek_grid[0] or Ek>Ek_grid[-1]:
        return 0
    #
    b_idx = int((Ek-Ek_grid[0])/(Ek_grid[1]-Ek_grid[0]))+1
    if b_idx == len(Ek_grid): b_idx -= 1
    a_idx = b_idx - 1
    b, a = Ek_grid[b_idx], Ek_grid[a_idx]
    #
    fb = hfk[l, b_idx]
    fa = hfk[l, a_idx]
    rez = (fb-fa)/(b-a)*Ek + (b*fa-a*fb)/(b-a)
    return pi*rez if eta+1 else pi*rez.conjugate()

def get_at_k1(Ek, Ek_grid, l, cb, conj, phi1k, hphi1k):
    """
    Linearly interpolate the values of phi1k, hphi1k on Ek_grid at point Ek.

    Parameters
    ----------
    Ek : float
        Energy value (not necessarily a grid point).
    Ek_grid : array
        Energy grid.
    l : int
        Lead label.
    cb : int
        Index corresponding to Phi[1](k) matrix element.
    conj : bool
        If conj=True the term in the integral equation is conjugated.
    phi1k : array
        Numpy array with dimensions (len(Ek_grid), nleads, ndm1, ndm0),
        cotaining energy resolved first order density matrix elements Phi[1](k).
    hphi1k : array
        Numpy array with dimensions (len(Ek_grid_ext), nleads, ndm1, ndm0),
        containing Hilbert transform of phi1k.

    Returns
    -------
    array, array
        Interpolated values of phi1k and hphi1k at Ek.
    """
    if Ek<Ek_grid[0] or Ek>Ek_grid[-1]:
        return 0, 0
    #
    #b_idx = np.searchsorted(Ek_grid, Ek)
    b_idx = int((Ek-Ek_grid[0])/(Ek_grid[1]-Ek_grid[0]))+1
    if b_idx == len(Ek_grid): b_idx -= 1
    a_idx = b_idx - 1
    b, a = Ek_grid[b_idx], Ek_grid[a_idx]
    #print(a_idx, b_idx, a, Ek, b)
    #
    fb = phi1k[b_idx, l, cb]
    fa = phi1k[a_idx, l, cb]
    u = (fb-fa)/(b-a)*Ek + (b*fa-a*fb)/(b-a)
    u = u.conjugate() if conj else u
    #
    fb = hphi1k[b_idx, l, cb]
    fa = hphi1k[a_idx, l, cb]
    hu = (fb-fa)/(b-a)*Ek + (b*fa-a*fb)/(b-a)
    hu = hu.conjugate() if conj else hu
    return pi*hu, pi*u

def phi1k_local_2vN(ind, Ek_grid, fk, hfkp, hfkm, E, Tba, si):
    """
    Constructs Phi[1](k) corresponding to local approximation at Ek_grid[ind].
    More precisely, generates local approximation kernels L1 and L0, Phi[1](k) = L0(k)Phi[0],
    L1(k)Phi[1](k) = L0p(k)Phi[0], L0 = L1^{-1}L0p.

    Parameters
    ----------
    ind : int
        Index of a point on a Ek_grid.
    Ek_grid : array
        Energy grid.
    fk : array
        nleads by len(Ek_grid) numpy array containing Fermi function.
    hfkp,hfkm : array
        Hilbert transforms of fk and 1-fk.
    E : array
        nmany by 1 array containing Hamiltonian eigenvalues.
    Tba : array
        nmany by nmany array, which contains many-body tunneling amplitude matrix.
    si : StateIndexingDM
        StateIndexingDM object.

    Returns
    -------
    kern0 : array
        Numpy array with dimensions (nleads, ndm1, ndm0).
        Gives local approximation kernel L0(k) at Ek_grid[ind].
        Shows how Phi[1](k) is expressed in terms of Phi[0].
    kern1 : array
        Numpy array with dimensions (nleads, ndm1, ndm1).
        Gives local approximation kernel L1(k) at Ek_grid[ind].
    """
    Ek = Ek_grid[ind]
    kern0 = np.zeros((si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
    kern1 = np.zeros((si.nleads, si.ndm1, si.ndm1), dtype=complexnp)
    for charge in range(si.ncharge-1):
        dcharge = charge+2
        ccharge = charge+1
        bcharge = charge
        acharge = charge-1
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = si.get_ind_dm1(c, b, bcharge)
            for l in range(si.nleads):
                # Note that the bias is put in the distributions and not in the dispersion
                kern1[l, cb, cb] += Ek-E[c]+E[b]
                fp = fk[l, ind] #fermi_func(+(Ek-mulst[l])/tlst[l])
                fm = 1-fp       #fermi_func(-(Ek-mulst[l])/tlst[l])
                # Phi[0] terms
                for b1 in si.statesdm[bcharge]:
                    b1b = si.get_ind_dm0(b1, b, bcharge)
                    kern0[l, cb, b1b] += +Tba[l, c, b1]*fp
                for c1 in si.statesdm[ccharge]:
                    cc1 = si.get_ind_dm0(c, c1, ccharge)
                    kern0[l, cb, cc1] += -Tba[l, c1, b]*fm
                #---------------------------------------------------------------------------
                # Phi[1] terms
                # 2nd and 7th terms
                for b1, a1 in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):
                    b1a1 = si.get_ind_dm1(b1, a1, acharge)
                    for l1 in range(si.nleads):
                        kern1[l, cb, b1a1] -= +Tba[l1, c, b1]*Tba[l1, a1, b]*(
                                                +func_2vN(+(Ek-E[b1]+E[b]), Ek_grid, l1, +1, hfkp)
                                                -func_2vN(-(Ek-E[c]+E[a1]), Ek_grid, l1, -1, hfkp) )
                # 6th and 8th terms
                for b1 in si.statesdm[bcharge]:
                    cb1 = si.get_ind_dm1(c, b1, bcharge)
                    for l1 in range(si.nleads):
                        for c1 in si.statesdm[ccharge]:
                            kern1[l, cb, cb1] -= ( +Tba[l1, b1, c1]*Tba[l1, c1, b]
                                                *func_2vN(+(Ek-E[c]+E[c1]), Ek_grid, l1, +1, hfkp) )
                        for a1 in si.statesdm[acharge]:
                            kern1[l, cb, cb1] -= ( -Tba[l1, b1, a1]*Tba[l1, a1, b]
                                                *func_2vN(-(Ek-E[c]+E[a1]), Ek_grid, l1, -1, hfkm) )
                # 1st and 3rd terms
                for c1 in si.statesdm[ccharge]:
                    c1b = si.get_ind_dm1(c1, b, bcharge)
                    for l1 in range(si.nleads):
                        for b1 in si.statesdm[bcharge]:
                            kern1[l, cb, c1b] -= ( +Tba[l1, c, b1]*Tba[l1, b1, c1]
                                                *func_2vN(+(Ek-E[b1]+E[b]), Ek_grid, l1, +1, hfkm) )
                        for d1 in si.statesdm[dcharge]:
                            kern1[l, cb, c1b] -= ( -Tba[l1, c, d1]*Tba[l1, d1, c1]
                                                *func_2vN(-(Ek-E[d1]+E[b]), Ek_grid, l1, -1, hfkp) )
                # 5th and 4th terms
                for d1, c1 in itertools.product(si.statesdm[dcharge], si.statesdm[ccharge]):
                    d1c1 = si.get_ind_dm1(d1, c1, ccharge)
                    for l1 in range(si.nleads):
                        kern1[l, cb, d1c1] -= +Tba[l1, c, d1]*Tba[l1, c1, b]*(
                                                +func_2vN(+(Ek-E[c]+E[c1]), Ek_grid, l1, +1, hfkm)
                                                -func_2vN(-(Ek-E[d1]+E[b]), Ek_grid, l1, -1, hfkm) )
    for l in range(si.nleads):
        kern0[l] = np.dot(np.linalg.inv(kern1[l]), kern0[l])
    return kern0, kern1

def phi1k_iterate_2vN(ind, Ek_grid, phi1k, hphi1k, fk, kern1, E, Tba, si):
    """
    Iterates the 2vN integral equation.

    Parameters
    ----------
    ind : int
        Index of a point on a Ek_grid.
    Ek_grid : array
        Energy grid.
    phi1k : array
        Numpy array with dimensions (len(Ek_grid), nleads, ndm1, ndm0)
        cotaining difference from phi1k after performing one iteration.
    hphi1k : array
        Numpy array with dimensions (len(Ek_grid), nleads, ndm1, ndm0)
        Hilbert transform of phi1k.
    fk : array
        nleads by len(Ek_grid) numpy array containing Fermi function.
    kern1 : array
        Local approximation kernel L1(k) at Ek_grid[ind].
    E : array
        nmany by 1 array containing Hamiltonian eigenvalues.
    Tba : array
        nmany by nmany array, which contains many-body tunneling amplitude matrix.
    si : StateIndexingDM
        StateIndexingDM object.

    Returns
    -------
    kern0 : array
        Numpy array with dimensions (nleads, ndm1, ndm0)
        Gives a correction to Phi[1](k) after iteration of integral equation.
        Shows how delta[Phi[1](k)] is expressed in terms of Phi[0].
    """
    Ek = Ek_grid[ind]
    kern0 = np.zeros((si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
    for charge in range(si.ncharge-1):
        dcharge = charge+2
        ccharge = charge+1
        bcharge = charge
        acharge = charge-1
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = si.get_ind_dm1(c, b, bcharge)
            for l in range(si.nleads):
                fp = fk[l, ind] #fermi_func((Ek-mulst[l])/tlst[l])
                fm = 1-fp
                term = np.zeros(si.ndm0, dtype=complexnp)
                # 1st term
                for a1 in si.statesdm[acharge]:
                    ba1 = si.get_ind_dm1(b, a1, acharge)
                    for b1, l1 in itertools.product(si.statesdm[bcharge], range(si.nleads)):
                        #print('1')
                        hu, u = get_at_k1(+(Ek-E[b1]+E[b]), Ek_grid, l1, ba1, True, phi1k, hphi1k)
                        term += -Tba[l1, c, b1]*Tba[l, b1, a1]*fp*(hu - 1j*u)
                # 2nd and 5th terms
                for b1, c1 in itertools.product(si.statesdm[bcharge], si.statesdm[ccharge]):
                    #print('2 and 5')
                    c1b1 = si.get_ind_dm1(c1, b1, bcharge)
                    for l1 in range(si.nleads):
                        # 2nd term
                        hu, u = get_at_k1(+(Ek-E[b1]+E[b]), Ek_grid, l1, c1b1, True, phi1k, hphi1k)
                        term += -Tba[l1, c, b1]*(hu - 1j*u)*fm*Tba[l, c1, b]
                        # 5th term
                        hu, u = get_at_k1(+(Ek-E[c]+E[c1]), Ek_grid, l1, c1b1, True, phi1k, hphi1k)
                        term += -Tba[l, c, b1]*fp*(hu - 1j*u)*Tba[l1, c1, b]
                # 3rd term
                for c1 in si.statesdm[ccharge]:
                    c1b = si.get_ind_dm1(c1, b, bcharge)
                    for d1, l1 in itertools.product(si.statesdm[dcharge], range(si.nleads)):
                        #print('3')
                        hu, u = get_at_k1(-(Ek-E[d1]+E[b]), Ek_grid, l1, c1b, False, phi1k, hphi1k)
                        term += +Tba[l1, c, d1]*Tba[l, d1, c1]*fp*(hu + 1j*u)
                # 4th term
                for d1, c1 in itertools.product(si.statesdm[dcharge], si.statesdm[ccharge]):
                    #print('4')
                    d1c1 = si.get_ind_dm1(d1, c1, ccharge)
                    for l1 in range(si.nleads):
                        hu, u = get_at_k1(-(Ek-E[d1]+E[b]), Ek_grid, l1, d1c1, False, phi1k, hphi1k)
                        term += +Tba[l1, c, d1]*(hu + 1j*u)*fm*Tba[l, c1, b]
                # 6th term
                for d1 in si.statesdm[dcharge]:
                    d1c = si.get_ind_dm1(d1, c, ccharge)
                    for c1, l1 in itertools.product(si.statesdm[ccharge], range(si.nleads)):
                        #print('6')
                        hu, u = get_at_k1(+(Ek-E[c]+E[c1]), Ek_grid, l1, d1c, True, phi1k, hphi1k)
                        term += -(hu - 1j*u)*fm*Tba[l, d1, c1]*Tba[l1, c1, b]
                # 7th term
                for b1, a1 in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):
                    #print('7')
                    b1a1 = si.get_ind_dm1(b1, a1, acharge)
                    for l1 in range(si.nleads):
                        hu, u = get_at_k1(-(Ek-E[c]+E[a1]), Ek_grid, l1, b1a1, False, phi1k, hphi1k)
                        term += +Tba[l, c, b1]*fp*(hu + 1j*u)*Tba[l1, a1, b]
                # 8th term
                for b1 in si.statesdm[bcharge]:
                    cb1 = si.get_ind_dm1(c, b1, bcharge)
                    for a1, l1 in itertools.product(si.statesdm[acharge], range(si.nleads)):
                        #print('8')
                        hu, u = get_at_k1(-(Ek-E[c]+E[a1]), Ek_grid, l1, cb1, False, phi1k, hphi1k)
                        term += +(hu + 1j*u)*fm*Tba[l, b1, a1]*Tba[l1, a1, b]
                kern0[l, cb] = term
    for l in range(si.nleads):
        kern0[l] = np.dot(np.linalg.inv(kern1[l]), kern0[l])
    return kern0

def get_phi1_phi0_2vN(sys):
    """
    Integrates phi1k over energy.

    Parameters
    ----------
    sys : Approach2vN
        Approach2vN object.

    Modifies:
    sys.phi1_phi0 : array
        Numpy array with dimensions (nleads, ndm1, ndm0),
        corresponding to energy integrated Phi[1](k).
    sys.e_phi1_phi0 : array
        Numpy array with dimensions (nleads, ndm1, ndm0),
        corresponding to energy integrated Ek*Phi[1](k).
    """
    #(phi1k, Ek_grid, si):
    (phi1k, Ek_grid, si) = (sys.phi1k, sys.Ek_grid, sys.si)
    # Get integrated Phi[1]_{cb} in terms of Phi[0]_{bb'}
    phi1_phi0 = np.zeros((si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
    e_phi1_phi0 = np.zeros((si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
    Eklen = len(Ek_grid)
    #
    for j1 in range(Eklen):
        # Trapezoidal rule
        if j1 == 0:         dx = Ek_grid[j1+1] - Ek_grid[j1]
        elif j1 == Eklen-1: dx = Ek_grid[j1]   - Ek_grid[j1-1]
        else:               dx = Ek_grid[j1+1] - Ek_grid[j1-1]
        phi1_phi0 += 1/2*dx*phi1k[j1]
        e_phi1_phi0 += 1/2*dx*Ek_grid[j1]*phi1k[j1]
    #
    sys.phi1_phi0 = phi1_phi0
    sys.e_phi1_phi0 = e_phi1_phi0
    return 0

def kern_phi0_2vN(sys):
    """
    From Phi[1](k) generate equations containing just Phi[0].

    Parameters
    ----------
    sys : Approach2vN
        Approach2vN object.

    Modifies:
    sys.kern : array
        Kernel for master equation involving just Phi[0].
    sys.bvec : array
        Right hand side column vector for master equation.
    """
    #(phi1_phi0, E, Tba, si, funcp, mulst, tlst, dlst)
    (phi1_phi0, E, Tba, si) = (sys.phi1_phi0, sys.qd.Ea, sys.leads.Tba, sys.si)
    symq = sys.funcp.symq
    norm_rowp = sys.funcp.norm_row
    # Integrated Phi[1]_{bc} in terms of phi1_phi0
    shuffle = np.zeros((si.ndm0, si.ndm0), dtype=intnp)
    for charge in range(si.ncharge):
        for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):
            bbp = si.get_ind_dm0(b, bp, charge)
            bpb = si.get_ind_dm0(bp, b, charge)
            shuffle[bbp, bpb] = 1
            shuffle[bpb, bbp] = 1
    phi1_phi0_conj = np.dot(np.conjugate(phi1_phi0), shuffle)
    # Set-up normalisation row
    norm_row = norm_rowp if symq else si.ndm0
    last_row = si.ndm0-1 if symq else si.ndm0
    bvec = np.zeros(last_row+1, dtype=complexnp)
    bvec[norm_row] = 1
    # Make equations for Phi[0]
    kern = np.zeros((last_row+1, si.ndm0), dtype=complexnp)
    norm_vec = np.zeros(si.ndm0, dtype=complexnp)
    for charge in range(si.ncharge):
        acharge = charge-1
        bcharge = charge
        ccharge = charge+1
        for b, bp in itertools.product(si.statesdm[bcharge], si.statesdm[bcharge]):
            bbp = si.get_ind_dm0(b, bp, bcharge)
            kern[bbp, bbp] = E[b]-E[bp]
            if b == bp:
                norm_vec[bbp] = 1
            #
            for a1 in si.statesdm[acharge]:
                bpa1 = si.get_ind_dm1(bp, a1, acharge)
                ba1 = si.get_ind_dm1(b, a1, acharge)
                for l in range(si.nleads):
                    kern[bbp] += +Tba[l, b, a1]*phi1_phi0_conj[l, bpa1]
                    kern[bbp] += -phi1_phi0[l, ba1]*Tba[l, a1, bp]
            #
            for c1 in si.statesdm[ccharge]:
                c1bp = si.get_ind_dm1(c1, bp, bcharge)
                c1b = si.get_ind_dm1(c1, b, bcharge)
                for l in range(si.nleads):
                    kern[bbp] += +Tba[l, b, c1]*phi1_phi0[l, c1bp]
                    kern[bbp] += -phi1_phi0_conj[l, c1b]*Tba[l, c1, bp]
    kern[norm_row] = norm_vec
    #
    sys.kern = kern
    sys.bvec = bvec
    return 0

def generate_current_2vN(sys):
    """
    Finds the currents for 2vN approach after an approximate
    solution for integral 2vN equations was found.

    Parameters
    ----------
    sys : Approach2vN
        Approach2vN object.

    Modifies:
    sys.phi1 : array
        nleads by ndm1 numpy array, which gives energy integrated
        Phi[1](k) after the values for Phi[0] were inserted.
    sys.current : array
        Values of the current having nleads entries.
    sys.energy_current : array
        Values of the energy current having nleads entries.
    sys.heat_current : array
        Values of the heat current having nleads entries.
    """
    #(phi1_phi0, e_phi1_phi0, phi0, Tba, si)
    (phi1_phi0, e_phi1_phi0) = (sys.phi1_phi0, sys.e_phi1_phi0)
    (phi0, Tba, si) = (sys.phi0, sys.leads.Tba, sys.si)
    #
    phi1 = np.zeros((si.nleads, si.ndm1), dtype=complexnp)
    h_phi1 = np.zeros((si.nleads, si.ndm1), dtype=complexnp)
    for l in range(si.nleads):
        phi1[l] = np.dot(phi1_phi0[l], phi0)
        h_phi1[l] = np.dot(e_phi1_phi0[l], phi0)
    #
    current = np.zeros(si.nleads, dtype=complexnp)
    energy_current = np.zeros(si.nleads, dtype=complexnp)
    for charge in range(si.ncharge-1):
        ccharge = charge+1
        bcharge = charge
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = si.get_ind_dm1(c, b, bcharge)
            for l in range(si.nleads):
                current[l] += phi1[l, cb]*Tba[l, b, c]
                energy_current[l] = energy_current[l] + h_phi1[l, cb]*Tba[l, b, c]
    sys.phi1 = phi1
    sys.current = np.array(-2*current.imag, dtype=doublenp)
    sys.energy_current = np.array(-2*energy_current.imag, dtype=doublenp)
    sys.heat_current = sys.energy_current - sys.current*sys.leads.mulst
    return 0

def get_emin_emax(sys):
    """
    Finds how much the Ek_grid has to be expanded
    above the bandwidht D of the leads.

    Parameters
    ----------
    sys : Approach2vN
        Approach2vN object.

    Modifies:
    sys.funcp.emin : float
        Minimal energy in the updated Ek_grid.
    sys.funcp.emax : float
        Maximal energy in the updated Ek_grid.
    """
    #(E, si, dband) = (sys.qd.Ea, sys.si, sys.leads.dlst[0,1])
    (E, si, dmin, dmax) = (sys.qd.Ea, sys.si, sys.funcp.dmin, sys.funcp.dmax)
    lst = [dmin, dmax]
    for charge in range(si.ncharge):
        for b, bp in itertools.product(si.statesdm[charge], si.statesdm[charge]):
            lst.append(dmax-E[b]+E[bp])
            lst.append(dmin-E[b]+E[bp])
    for charge in range(si.ncharge-2):
        for d, b in itertools.product(si.statesdm[charge+2], si.statesdm[charge]):
            lst.append(dmin+E[d]-E[b])
            lst.append(dmax+E[d]-E[b])
    sys.funcp.emax = max(lst)
    sys.funcp.emin = min(lst)
    return 0

def get_grid_ext(sys):
    """
    Expands the Ek_grid from emin to emax.

    Parameters
    ----------
    sys : Approach2vN
        Approach2vN object.

    Modifies:
    sys.Ek_grid_ext : array
        Extended Ek_grid from emin to emax.
    sys.funcp.kpnt_left : int
        Number of points Ek_grid is extended to the left.
    sys.funcp.kpnt_right : int
        Number of points Ek_grid is extended to the right.
    """
    (emin_, emax_, dmin, dmax, ext_fct) = (sys.funcp.emin, sys.funcp.emax,
                                           sys.funcp.dmin, sys.funcp.dmax,
                                           sys.funcp.ext_fct)
    Ek_grid = sys.Ek_grid
    step = Ek_grid[1]-Ek_grid[0]
    emin = ext_fct*(emin_-dmin)+dmin
    emax = ext_fct*(emax_-dmax)+dmax
    ext_left = np.sort(-np.arange(-dmin+step, -emin+step, step))
    ext_right = np.arange(dmax+step, emax+step, step)
    sys.Ek_grid_ext = np.concatenate((ext_left, Ek_grid, ext_right))
    sys.funcp.kpnt_left, sys.funcp.kpnt_right = len(ext_left), len(ext_right)
    return 0

def get_htransf_phi1k(phi1k, funcp):
    """
    Performs Hilbert transform of phi1k.

    Parameters
    ----------
    phi1k : array
        Numpy array with dimensions (len(Ek_grid), nleads, ndm1, ndm0)
        cotaining energy resolved first order density matrix elements Phi[1](k).
    funcp : FunctionProperties
        FunctionProperties object.

    Modifies:
    funcp.ht_ker : array
        Kernel used when performing Hilbert transform using FFT.

    Returns
    -------
    phi1k : array
        Numpy array with dimensions (len(Ek_grid_ext), nleads, ndm1, ndm0).
        phi1k padded from the left and the right by zeros.
    hphi1k : array
        Numpy array with dimensions (len(Ek_grid_ext), nleads, ndm1, ndm0).
        Hilbert transform of phi1k on extended grid.

    """
    nleads, ndm1, ndm0 = phi1k.shape[1], phi1k.shape[2], phi1k.shape[3]
    Eklen_ext = phi1k.shape[0] + funcp.kpnt_left + funcp.kpnt_right
    # Create the kernels for Hilbert transformation
    if funcp.ht_ker is None or 2*Eklen_ext != len(funcp.ht_ker):
        funcp.ht_ker = kernel_fredriksen(Eklen_ext)
    # Make phi1k on extended grid Ek_grid_ext
    # Pad phi1k values with zeros from the left and the right
    phi1k = np.concatenate( (np.zeros((funcp.kpnt_left, nleads, ndm1, ndm0)),
                             phi1k,
                             np.zeros((funcp.kpnt_right, nleads, ndm1, ndm0))), axis=0)
    # Make the Hilbert transformation
    hphi1k = np.zeros(phi1k.shape, dtype=complexnp)
    for l, cb, bbp in itertools.product(range(nleads), range(ndm1), range(ndm0)):
        #print(l, cb, bbp)
        hphi1k[:, l, cb, bbp] = hilbert_fredriksen(phi1k[:, l, cb, bbp], funcp.ht_ker)
    return phi1k, hphi1k

def get_htransf_fk(fk, funcp):
    """
    Performs Hilbert transform of Fermi function.
    The energy is shifted by positive infinitesimal
    to the complex upper half-plane.

    Parameters
    ----------
    fk : array
        Numpy array with dimensions (nleads, len(Ek_grid)),
        cotaining Fermi function values on the grid Ek_grid.
    funcp : FunctionProperties
        FunctionProperties object.

    Modifies:
    funcp.ht_ker : array
        Kernel used when performing Hilbert transform using FFT.

    Returns
    -------
    fk : array
        Numpy array with dimensions (nleads, len(Ek_grid_ext)).
        fk padded from the left and the right by zeros.
    hfk : array
        Numpy array with dimensions (nleads, len(Ek_grid_ext)).
        Hilbert transform of fk on extended grid.
    """
    nleads = fk.shape[0]
    Eklen_ext = fk.shape[1] + funcp.kpnt_left + funcp.kpnt_right
    # Create the kernels for Hilbert transformation
    if funcp.ht_ker is None or 2*Eklen_ext != len(funcp.ht_ker):
        funcp.ht_ker = kernel_fredriksen(Eklen_ext)
    # Pad fk values with zeros from the left and the right
    fk = np.concatenate( (np.zeros((nleads, funcp.kpnt_left)),
                          fk,
                          np.zeros((nleads, funcp.kpnt_right))), axis=1)
    # Calculate the Hilbert transform with added positive infinitesimal of the Fermi functions
    hfk = np.zeros((nleads, Eklen_ext), dtype=doublenp)
    for l in range(nleads):
        hfk[l] = hilbert_fredriksen(fk[l], funcp.ht_ker).real
    # The energy is shifted by positive infinitesimal to the complex upper half-plane
    hfk = hfk - 1j*fk
    return fk, hfk

def iterate_2vN(sys):
    """
    Performs the iterative solution of the 2vN approach integral equations.

    Parameters
    ----------
    sys : Approach2vN
        Approach2vN object.

    Modifies:
    sys.phi1k_delta : array
        Numpy array with dimensions (len(Ek_grid), nleads, ndm1, ndm0).
        Correction to Phi[1](k) after an iteration.
    sys.hphi1k_delta : array
        Numpy array with dimensions (len(Ek_grid_ext), nleads, ndm1, ndm0).
        Hilbert transform of phi1k_delta on extended grid Ek_grid_ext.
    sys.kern1k : array
        Numpy array with dimensions (len(Ek_grid), nleads, ndm1, ndm1)
        corresponding to energy resolved local kernel for Phi[1](k).
    sys.funcp.emin : float
        Minimal energy in the updated Ek_grid.
    sys.funcp.emax : float
        Maximal energy in the updated Ek_grid.
    sys.Ek_grid_ext : array
        Extended Ek_grid from emin to emax.
    sys.funcp.kpnt_left : int
        Number of points Ek_grid is extended to the left.
    sys.funcp.kpnt_right : int
        Number of points Ek_grid is extended to the right.
    sys.fkp : array
        nleads by len(Ek_grid_ext) numpy array containing Fermi function.
    sys.fkm : array
        nleads by len(Ek_grid_ext) numpy array containing 1-Fermi function.
    sys.hfkp : array
        Hilbert transform of fkp.
    sys.hfkm : array
        Hilbert transform of fkm.
    """
    (Ek_grid, Ek_grid_ext) = (sys.Ek_grid, sys.Ek_grid_ext)
    (E, Tba, si, funcp) = (sys.qd.Ea, sys.leads.Tba, sys.si, sys.funcp)
    (mulst, tlst) = (sys.leads.mulst, sys.leads.tlst)
    # Assign sys.phi1k_delta to phi1k_delta_old, because the new phi1k_delta
    # will be generated in this function
    (phi1k_delta_old, kern1k) = (sys.phi1k_delta, sys.kern1k)
    #
    Eklen = len(Ek_grid)
    if phi1k_delta_old is None:
        ## Define the extended grid Ek_grid_ext for calculations outside the bandwidth
        # Here sys.funcp.emin, sys.funcp.emax are defined
        get_emin_emax(sys)
        # Here sys.Ek_grid_ext, sys.funcp.kpnt_left, sys.funcp.kpnt_right are defined
        get_grid_ext(sys)
        Ek_grid_ext = sys.Ek_grid_ext
        Eklen_ext = len(Ek_grid_ext)
        # Generate the Fermi functions on the grid
        # This is necessary to generate only if Ek_grid, mulst, or tlst are changed
        sys.fkp = np.zeros((si.nleads, Eklen), dtype=doublenp)
        for l in range(si.nleads):
            sys.fkp[l] = 1/( np.exp((Ek_grid - mulst[l])/tlst[l]) + 1 )
        sys.fkm = 1-sys.fkp
        sys.fkp, sys.hfkp = get_htransf_fk(sys.fkp, funcp)
        sys.fkm, sys.hfkm = get_htransf_fk(sys.fkm, funcp)
        # Calculate the zeroth iteration of Phi[1](k)
        phi1k_delta = np.zeros((Eklen, si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
        kern1k = np.zeros((Eklen, si.nleads, si.ndm1, si.ndm1), dtype=complexnp)
        for j1 in range(Eklen):
            ind = j1 + funcp.kpnt_left
            phi1k_delta[j1], kern1k[j1] = phi1k_local_2vN(ind, Ek_grid_ext, sys.fkp,
                                                          sys.hfkp, sys.hfkm, E, Tba, si)
        hphi1k_delta = None
    elif kern1k is None:
        pass
    else:
        # Hilbert transform phi1k_delta_old on extended grid Ek_grid_ext
        #print('Hilbert transforming')
        phi1k_delta_old, hphi1k_delta = get_htransf_phi1k(phi1k_delta_old, funcp)
        #print('Making an iteration')
        phi1k_delta = np.zeros((Eklen, si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
        for j1 in range(Eklen):
            ind = j1 + funcp.kpnt_left
            phi1k_delta[j1] = phi1k_iterate_2vN(ind, Ek_grid_ext, phi1k_delta_old, hphi1k_delta,
                                                sys.fkp, kern1k[j1], E, Tba, si)
    #
    sys.phi1k_delta = phi1k_delta
    sys.hphi1k_delta = hphi1k_delta
    sys.kern1k = kern1k
    return 0

class Approach_py2vN(Approach2vN):

    kerntype = 'py2vN'
    iterate = staticmethod(iterate_2vN)
    get_phi1_phi0 = staticmethod(get_phi1_phi0_2vN)
    kern_phi0 = staticmethod(kern_phi0_2vN)
    generate_current = staticmethod(generate_current_2vN)



"""Module for defining types used in neumann package."""

import numpy as np

boolnp = np.uint8          #np.bool
intnp = np.int
longnp = np.int64          #np.long
doublenp = np.float64      #np.double
complexnp = np.complex128



"""Module containing various functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from .indexing import szrange
from .indexing import ssqrange
from .indexing import sz_to_ind
from .indexing import ssq_to_ind

def get_charge(sys, b):
    '''
    Get charge of the many-body state b

    Parameters
    ----------
    sys : Approach, Approach2vN
        The system given as Approach or Approach2vN object.
    b : int
        Label of the many-body state.

    Returns
    -------
    int
        Charge of the many-body state b.
    '''
    return sum(sys.si.get_state(b))

def multiarray_sort(arr, srt=[0]):
    '''
    Sort rows of a two-dimensional array for a given
    hierarchy of rows.

    Parameters
    ----------
    arr : array
        A two-dimensional numpy array.
    srt : list
        List specifying in which order of rows to sort.

    Returns
    -------
    array
        A sorted array.
    '''
    ind = np.lexsort([arr[i] for i in reversed(srt)])
    return (arr.T[ind]).T

def sort_eigenstates(sys, srt='n'):
    '''
    Sort many-body states of the system by given order of properties.

    Parameters
    ----------
    sys : Approach, Approach2vN, or Builder
        The system given as Approach, Approach2vN, or Builder object.
    srt : list
        List specifying in which order of properties to sort.
        For example, in the case of 'ssq' indexing we have  such convention:
        0 - energy
        1 - charge
        2 - spin projection :math:`S_{z}`
        3 - total spin :math:`S^{2}`
        The default sorting order for 'ssq' indexing is srt=[1, 2, 3, 0]

    Modifies:
    sys.qd.Ea_ext : array
        A two-dimensional numpy array containing in the zeroth row energies,
        first row charge, and etc.
    sys.si.states_order : array
        A numpy row containing a new order of many-body states.
    '''
    if srt != 'n':
        if sys.qd.Ea_ext is None:
            sys.qd.Ea_ext = construct_Ea_extended(sys)
        else:
            sys.qd.Ea_ext[0] = sys.qd.Ea
        srt.append(-1)
        sys.si.states_order = np.array( multiarray_sort(sys.qd.Ea_ext, srt)[-1], dtype=int)
    else:
        sys.si.states_order = range(sys.si.nmany)

def get_phi0(sys, b_, bp_):
    '''
    Get the reduced density matrix element corresponding to
    many-body states b and bp.

    Parameters
    ----------
    sys : Approach, Approach2vN, or Builder
        The system given as Approach, Approach2vN, or Builder object.
    b\_,bp\_ : int
        Labels of the many-body states.

    Returns
    --------
    phi0bbp : complex
        A matrix element of the reduced density matrix (complex number).
    '''
    b = sys.si.states_order[b_]
    bp = sys.si.states_order[bp_]
    bcharge = sum(sys.si.get_state(b))
    bpcharge = sum(sys.si.get_state(bp))
    phi0bbp = 0.0
    if sys.funcp.kerntype == 'Pauli':
        if b == bp:
            ind = sys.si.get_ind_dm0(b, b, bcharge, maptype=1)
            phi0bbp = sys.phi0[ind]
    elif bcharge == bpcharge:
        ind = sys.si.get_ind_dm0(b, bp, bcharge, maptype=1)
        conj = sys.si.get_ind_dm0(b, bp, bcharge, maptype=3)
        if ind != -1:
            if type(sys.si).__name__ == 'StateIndexingDMc':
                phi0bbp = sys.phi0[ind]
            else:
                ndm0, npauli = sys.si.ndm0, sys.si.npauli
                phi0bbp = ( sys.phi0[ind] + 1j*sys.phi0[ndm0-npauli+ind]
                                              * (+1 if conj else -1)
                                              * (0 if ind < npauli else 1) )
    return phi0bbp

def get_phi1(sys, l, c_, b_):
    '''
    Get the energy integrated current amplitudes corresponding to
    lead l and many-body states c and b.

    Parameters
    ----------
    sys : Approach, Approach2vN, or Builder
        The system given as Approach, Approach2vN, or Builder object.
    l : int
        Label of the lead channel.
    c\_,b\_ : int
        Labels of the many-body states.

    Returns
    --------
    phi0bbp : complex
        A matrix element of the reduced density matrix (complex number).
    '''
    if sys.funcp.kerntype == 'Pauli':
        return None
    else:
        c = sys.si.states_order[c_]
        b = sys.si.states_order[b_]
        ccharge = sum(sys.si.get_state(c))
        bcharge = sum(sys.si.get_state(b))
        phi1cb = 0.0
        if ccharge == bcharge+1:
            ind = sys.si.get_ind_dm1(c, b, bcharge)
            phi1cb = sys.phi1[l, ind]
        elif ccharge+1 == bcharge:
            ind = sys.si.get_ind_dm1(b, c, ccharge)
            phi1cb = sys.phi1[l, ind].conjugate()
        return phi1cb

def construct_Ea_extended(sys):
    '''
    Constructs an array containing properties of the many-body states,
    like energy, charge, spin-projection :math:`S_{z}`, etc.


    Parameters
    ----------
    sys : Approach, Approach2vN, or Builder
        The system given as Approach, Approach2vN, or Builder object.

    Returns
    --------
    Ea_ext : array
        | A two-dimensional numpy array containing in the zeroth row energies,
          first row charge, and etc. We have such convention:
          Ea_ext[property number, state number]
        |
        | For example, for 'charge' indexing we have such properties:
        | 0 - energy
        | 1 - charge
        | 2 - state number
        |
        | and for 'ssq' indexing we have:
        | 0 - energy
        | 1 - charge
        | 2 - spin projection :math:`S_{z}`
        | 3 - total spin :math:`S^{2}`
        | 4 - state number
    '''
    (si, Ea) = (sys.si, sys.qd.Ea)
    if si.indexing == 'sz':
        Ea_ext = np.zeros((4, len(Ea)), dtype=float)
        Ea_ext[0] = sys.qd.Ea
        Ea_ext[3] = np.arange(si.nmany)
        # Iterate over charges
        for charge in range(si.ncharge):
            # Iterate over spin projection sz
            for sz in szrange(charge, si.nsingle):
                # Iterate over many-body states for given charge and sz
                szind = sz_to_ind(sz, charge, si.nsingle)
                for ind in range(len(si.szlst[charge][szind])):
                    # The mapping of many-body states is according to szlst
                    sn = si.szlst[charge][szind][ind]
                    Ea_ext[1,sn] = charge
                    Ea_ext[2,sn] = sz
    elif si.indexing == 'ssq':
        Ea_ext = np.zeros((5, len(Ea)), dtype=float)
        Ea_ext[0] = sys.qd.Ea
        Ea_ext[4] = np.arange(si.nmany)
        # Iterate over charges
        for charge in range(si.ncharge):
            # Iterate over spin projection sz
            for sz in szrange(charge, si.nsingle):
                szind = sz_to_ind(sz, charge, si.nsingle)
                # Iterate over total spin ssq
                for ssq in ssqrange(charge, sz, si.nsingle):
                    ssqind = ssq_to_ind(ssq, sz)
                    # Iterate over many-body states for given charge, sz, and ssq
                    for ind in range(len(si.ssqlst[charge][szind][ssqind])):
                        # The mapping of many-body states is according to ssqlst
                        sn = si.ssqlst[charge][szind][ssqind][ind]
                        Ea_ext[1,sn] = charge
                        Ea_ext[2,sn] = sz
                        Ea_ext[3,sn] = ssq
    else:
        # Iterate over charges
        Ea_ext = np.zeros((3, len(Ea)), dtype=float)
        Ea_ext[0] = sys.qd.Ea
        Ea_ext[2] = np.arange(si.nmany)
        for charge in range(si.ncharge):
            # Iterate over many-body states for given charge
            for ind in range(len(si.chargelst[charge])):
                # The mapping of many-body states is according to chargelst
                sn = si.chargelst[charge][ind]
                Ea_ext[1,sn] = charge
    return Ea_ext

def remove_coherences(sys, dE):
    '''
    Remove the coherences with energy difference larger than dE.

    Parameters
    ----------
    sys : Approach, Approach2vN, or Builder
        The system given as Approach, Approach2vN, or Builder object.
    dE : float
        Energy difference.

    Modifies:
    sys.si.mapdm0 : list
        List showing which density matrix elements are mapped to each other due to symmetries
        and which density matrix elements are neglected (entries with values -1).
    '''
    # Find which coherences to remove
    si, E = sys.si, sys.qd.Ea
    ilst, indlst = [], []
    for i in range(si.ndm0_):
        if si.mapdm0[i] != -1:
            ind = si.mapdm0[i]
            b, bp = si.inddm0[ind]
            if abs(E[b]-E[bp]) > dE:
                si.mapdm0[i] = -1
            else:
                ilst.append(i)
                indlst.append(ind)
    # Count the new number of density matrix elements
    srt = multiarray_sort(np.array([ilst, indlst]), [1,0])
    ilst, indlst = srt[0], srt[1]
    ind, count = 0, 0
    for i in range(len(indlst)):
        if indlst[i] == ind:
            indlst[i] = count
        else:
            count += 1
            ind = indlst[i]
            indlst[i] = count
    # Relabel the density matrix elements
    si.ndm0 = count+1
    si.ndm0r = 2*si.ndm0-si.npauli
    for i in range(len(indlst)):
        si.mapdm0[ilst[i]] = indlst[i]

def remove_states(sys, dE):
    '''
    Remove the states with energy dE larger than the ground state
    for the transport calculations.

    Parameters
    ----------
    sys : Approach, Approach2vN, or Builder
        The system given as Approach, Approach2vN, or Builder object.
    dE : float
        Energy above the ground state.

    Modifies:
    sys.si.statesdm : list
        List containing indices of many-body state under consideration.
    '''
    Emax = min(sys.qd.Ea)+dE
    statesdm = [[] for i in range(sys.si.nsingle+1)]
    for charge in range(sys.si.ncharge):
        for b in sys.si.chargelst[charge]:
            if sys.qd.Ea[b] < Emax:
                statesdm[charge].append(b)
    sys.si.set_statesdm(statesdm)

def use_all_states(sys):
    '''
    Use all states for the transport calculations.

    Parameters
    ----------
    sys : Approach, Approach2vN, or Builder
        The system given as Approach, Approach2vN, or Builder object.

    Modifies:
    sys.si.statesdm : list
        List containing indices of many-body state under consideration.
    '''
    sys.si.set_statesdm(sys.si.chargelst)

def print_state(sys, b_, eps=0.0, prntq=True, filename=None, separator=''):
    '''
    Prints properties of given many-body eigenstate of the quantum dot Hamiltonain

    Parameters
    ----------
    sys : Approach, Approach2vN, or Builder
        The system given as Approach, Approach2vN, or Builder object.
    b_ : int
        Index of the many-body eigenstate.
    eps : float
        Value at which expansion coefficients of the many-body eigenstate in Fock basis
        are neglected.
    prntq : bool
        If true then eigenstate properties are printed into command line.
    filename : string
        File to which to print eigenstate properties. The output is appended to this file.
    separator : string
        String added at the begining of the output.
    '''
    si = sys.si
    b = si.states_order[b_]
    Eb = sys.qd.Ea[b]
    sout = ('Energy: '+ str(Eb)+'\n'
           +'Eigenstate coefficients:')
    #
    if si.indexing == 'sz':
        charge, sz, alpha = si.ind_qn[b]
        sout = 'S^z [in units of 1/2]: '+ str(sz) + '\n' + sout
        szind = sz_to_ind(sz, charge, si.nsingle)
        ind = si.szlst[charge][szind].index(b)
        coeffs = np.array([abs(sys.qd.vecslst[charge][szind][:, ind]),
                           range(len(si.szlst[charge][szind]))])
        coeffs = np.array(multiarray_sort(coeffs)[1], dtype=int)
        #for j1 in range(len(si.szlst[charge][szind])):
        for j1 in reversed(coeffs):
            sn = si.szlst[charge][szind][j1]
            val = sys.qd.vecslst[charge][szind][j1, ind]
            if abs(val) >= eps:
                sout = sout + '\n'+'    |'+si.get_state(sn, strq=True)+'>: ' + str(val)
    elif si.indexing == 'ssq':
        charge, sz, ssq, alpha = si.ind_qn[b]
        sout = ('S^z [in units of 1/2]: ' + str(sz) + '\n'
               +'S^2 [in units of 1/2]: ' + str(ssq) + '\n' + sout)
        szind = sz_to_ind(sz, charge, si.nsingle)
        ssqind = ssq_to_ind(ssq, sz)
        ind = si.ssqlst[charge][szind][ssqind].index(b)
        coeffs = np.array([abs(sys.qd.vecslst[charge][szind][ssqind][:, ind]),
                           range(len(si.szlst[charge][szind]))])
        coeffs = np.array(multiarray_sort(coeffs)[1], dtype=int)
        #for j1 in range(len(si.szlst[charge][szind])):
        for j1 in reversed(coeffs):
            sn = si.szlst[charge][szind][j1]
            val = sys.qd.vecslst[charge][szind][ssqind][j1, ind]
            if abs(val) >= eps:
                sout = sout + '\n'+'    |'+si.get_state(sn, strq=True)+'>: ' + str(val)
    else:
        charge, alpha = si.ind_qn[b]
        #ind = si.dictdm[b]
        ind = si.chargelst[charge].index(b)
        coeffs = np.array([abs(sys.qd.vecslst[charge][:, ind]),
                           range(len(si.chargelst[charge]))])
        coeffs = np.array(multiarray_sort(coeffs)[1], dtype=int)
        #for j1 in range(len(si.chargelst[charge])):
        for j1 in reversed(coeffs):
            sn = si.chargelst[charge][j1]
            val = sys.qd.vecslst[charge][j1, ind]
            if abs(val) >= eps:
                sout = sout + '\n'+'    |'+si.get_state(sn, strq=True)+'>: ' + str(val)
    #
    sout = (separator
           +'Sorted label: ' + str(b_) + '\n'
           +'Original label: ' + str(b) + '\n'
           +'Charge: ' + str(charge) + '\n' + sout)
    #
    if not filename is None:
        with open(filename, 'a') as f:
            f.writelines(sout+'\n')
    if prntq:
        print(sout)
    pass

def print_all_states(sys, filename, eps=0.0, separator='', mode='w'):
    '''
    Prints properties of all many-body eigenstates to a file.

    Parameters
    ----------
    sys : Approach, Approach2vN, or Builder
        The system given as Approach, Approach2vN, or Builder object.
    filename : string
        File to which to print eigenstate properties. The output is appended to this file.
    eps : float
        Value at which expansion coefficients of the many-body eigenstate in Fock basis
        are neglected.
    separator : string
        String added at the begining of the output.
    mode : string
        Mode in which the file will be treated. For example, to append use 'a' or
        to write into clea file use 'w'.
    '''
    with open(filename, mode) as f:
        f.close()
    for ind in range(sys.si.nmany):
        print_state(sys, ind, eps, False, filename, separator)
    pass



"""Module for defining tunneling amplitudes from quantum dot to leads."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools
import numbers

import numpy as np

from .indexing import szrange
from .indexing import ssqrange
from .indexing import sz_to_ind
from .indexing import ssq_to_ind

from .mytypes import doublenp

def construct_Tba(leads, tleads, Tba_=None):
    """
    Constructs many-body tunneling amplitude matrix Tba from single particle
    tunneling amplitudes.

    Parameters
    ----------
    leads : LeadsTunneling
        LeadsTunneling object.
    tleads : dict
        Dictionary containing single particle tunneling amplitudes.
        tleads[(lead, state)] = tunneling amplitude.
    Tba_ : None or array
        nbaths by nmany by nmany numpy array containing old values of Tba.
        The values in tleads are added to Tba\_.

    Returns
    -------
    Tba : array
        nleads by nmany by nmany numpy array containing many-body tunneling amplitudes.
        The returned Tba corresponds to Fock basis.
    """
    si, mtype = leads.si, leads.mtype
    if Tba_ is None:
        Tba = np.zeros((si.nleads, si.nmany, si.nmany), dtype=mtype)
    else:
        Tba = Tba_
    # Iterate over many-body states
    for j1 in range(si.nmany):
        state = si.get_state(j1)
        # Iterate over single particle states
        for j0 in tleads:
            (j3, j2), tamp = j0, tleads[j0]
            # Calculate fermion sign for added/removed electron in a given state
            fsign = np.power(-1, sum(state[0:j2]))
            if state[j2] == 0:
                statep = list(state)
                statep[j2] = 1
                ind = si.get_ind(statep)
                Tba[j3, ind, j1] += fsign*tamp
            else:
                statep = list(state)
                statep[j2] = 0
                ind = si.get_ind(statep)
                Tba[j3, ind, j1] += fsign*np.conj(tamp)
    return Tba

def construct_full_pmtr(vecslst, si, mtype=complex):
    """
    From definite charge eigenvectors constructs full eigenvectors
    defined by indexing in si.

    Parameters
    ----------
    veclst : list of arrays
        List of size ncharge containing arrays defining eigenvector matrices for given charge.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.
    mtype : type
        Defines type of pmtr matrix. For example, float, complex, etc.

    Returns
    -------
    pmtr : array
        nmany by nmany numpy array containing many-body eigenvectors.
    """
    pmtr = np.zeros((si.nmany, si.nmany), dtype=mtype)
    if si.indexing == 'sz':
        # Iterate over charges
        for charge in range(si.ncharge):
            # Iterate over spin projection sz
            for sz in szrange(charge, si.nsingle):
                szind = sz_to_ind(sz, charge, si.nsingle)
                # Iterate over many-body states for given charge
                for j1, j2 in itertools.product(range(len(si.szlst[charge][szind])),
                                                range(len(si.szlst[charge][szind]))):
                    (pmtr[si.szlst[charge][szind][j1], si.szlst[charge][szind][j2]]) = (
                        vecslst[charge][szind][j1, j2] )
    elif si.indexing == 'ssq':
        # Iterate over charges
        for charge in range(si.ncharge):
            # Iterate over spin projection sz
            for sz in szrange(charge, si.nsingle):
                szind = sz_to_ind(sz, charge, si.nsingle)
                # Iterate over total spin ssq
                for ssq in ssqrange(charge, sz, si.nsingle):
                    ssqind = ssq_to_ind(ssq, sz)
                    # Iterate over many-body states for given charge
                    for j1, j2 in itertools.product(range(len(si.szlst[charge][szind])),
                                                    range(len(si.ssqlst[charge][szind][ssqind]))):
                        (pmtr[si.szlst[charge][szind][j1], si.ssqlst[charge][szind][ssqind][j2]]) = (
                            vecslst[charge][szind][ssqind][j1, j2] )
    else:
        # Iterate over charges
        for charge in range(si.ncharge):
            # Iterate over many-body states for given charge
            for j1, j2 in itertools.product(range(len(si.chargelst[charge])),
                                            range(len(si.chargelst[charge]))):
                pmtr[si.chargelst[charge][j1], si.chargelst[charge][j2]] = vecslst[charge][j1, j2]
    return pmtr

def rotate_Tba(Tba0, vecslst, si, indexing='n', mtype=complex):
    """
    Rotates tunneling amplitude matrix Tba0 in Fock basis to Tba,
    which is in eigenstate basis of the quantum dot.

    Parameters
    ----------
    Tba0 : array
        nleads by nmany by nmany numpy array, giving tunneling amplitudes in Fock basis.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.
    indexing : string
        Specifies what kind of rotation procedure to use. Default is si.indexing.
    mtype : type
        Defines type of Tba matrix. For example, float, complex, etc.

    Returns
    -------
    Tba : array
        nleads by nmany by nmany numpy array containing many-body tunneling amplitudes.
        The returned Tba corresponds to the quantum dot eigenbasis.
    """
    if indexing == 'n':
        indexingp = si.indexing
    else:
        indexingp = indexing
    Tba = np.zeros((si.nleads, si.nmany, si.nmany), dtype=mtype)
    if indexingp == 'Lin':
        pmtr = construct_full_pmtr(vecslst, si, mtype)
        for l in range(si.nleads):
            # Calculate many-body tunneling matrix Tba=P^(-1).Tba0.P
            # in eigenbasis of Hamiltonian from tunneling matrix Tba0 in Fock basis.
            # pmtr.conj().T denotes the conjugate transpose of pmtr.
            Tba[l] = np.dot(pmtr.conj().T, np.dot(Tba0[l], pmtr))
    elif indexingp == 'sz':
        for l, charge in itertools.product(range(si.nleads), range(si.ncharge-1)):
            szrng = szrange(charge, si.nsingle)
            # Lead labels from 0 to nleads//2 correspond to spin up
            # and nleads//2+1 to nleads-1 correspond to spin down
            if charge >= si.ncharge//2:
                szrng = szrng[0:-1] if l < si.nleads//2 else szrng[1:]
            # s=+1/-1 add/remove spin up/donw
            s = +1 if l < si.nleads//2 else -1
            for sz in szrng:
                szind = sz_to_ind(sz, charge, si.nsingle)
                szind2 = sz_to_ind(sz+s, charge+1, si.nsingle)
                i1 = si.szlst[charge][szind][0]
                i2 = si.szlst[charge][szind][-1] + 1
                i3 = si.szlst[charge+1][szind2][0]
                i4 = si.szlst[charge+1][szind2][-1] + 1
                Tba[l, i1:i2][:, i3:i4] = np.dot(vecslst[charge][szind].conj().T,
                                                 np.dot(Tba0[l, i1:i2][:, i3:i4],
                                                        vecslst[charge+1][szind2]))
                Tba[l, i3:i4][:, i1:i2] = Tba[l, i1:i2][:, i3:i4].conj().T
    elif indexingp == 'ssq':
        for l, charge in itertools.product(range(si.nleads), range(si.ncharge-1)):
            szrng = szrange(charge, si.nsingle)
            # Lead labels from 0 to nleads//2 correspond to spin up
            # and nleads//2+1 to nleads-1 correspond to spin down
            if charge >= si.ncharge//2:
                szrng = szrng[0:-1] if l < si.nleads//2 else szrng[1:]
            # s=+1/-1 add/remove spin up/donw
            s = +1 if l < si.nleads//2 else -1
            for sz in szrng:
                szind = sz_to_ind(sz, charge, si.nsingle)
                szind2 = sz_to_ind(sz+s, charge+1, si.nsingle)
                i1 = si.szlst[charge][szind][0]
                i2 = si.szlst[charge][szind][-1] + 1
                i3 = si.szlst[charge+1][szind2][0]
                i4 = si.szlst[charge+1][szind2][-1] + 1
                vecslst1 = np.concatenate(vecslst[charge][szind], axis=1)
                vecslst2 = np.concatenate(vecslst[charge+1][szind2], axis=1)
                Tba[l, i1:i2][:, i3:i4] = np.dot(vecslst1.conj().T,
                                                 np.dot(Tba0[l, i1:i2][:, i3:i4],
                                                        vecslst2))
                Tba[l, i3:i4][:, i1:i2] = Tba[l, i1:i2][:, i3:i4].conj().T
    elif indexingp == 'charge':
        for l, charge in itertools.product(range(si.nleads), range(si.ncharge-1)):
            i1 = si.chargelst[charge][0]
            i2 = si.chargelst[charge][-1] + 1
            i3 = si.chargelst[charge+1][0]
            i4 = si.chargelst[charge+1][-1] + 1
            Tba[l, i1:i2][:, i3:i4] = np.dot(vecslst[charge].conj().T,
                                             np.dot(Tba0[l, i1:i2][:, i3:i4],
                                                    vecslst[charge+1]))
            Tba[l, i3:i4][:, i1:i2] = Tba[l, i1:i2][:, i3:i4].conj().T
    return Tba
#---------------------------------------------------------------------------------------------------

def make_tleads_mtr(tleads, nleads, nsingle, mtype=complex):
    """
    Makes single particle tunneling matrix from a list or a dictionary.

    Parameters
    ----------
    tleads : list or dictionary
        Contains single particle tunneling amplitudes
    nleads : int
        Number of leads.
    nsingle : int
        Number of single particle states.
    mtype : type
        Defines type of tleads matrix. For example, float, complex, etc.

    Returns
    -------
    tleads_mtr : array
        nleads by nsingle numpy array containing single particle tunneling amplitudes.
    """
    tleads_mtr = np.zeros((nleads, nsingle), dtype=mtype)
    for j0 in tleads:
        if isinstance(tleads, list):    j1, j2, tamp = j0
        elif isinstance(tleads, dict): (j1, j2), tamp = j0, tleads[j0]
        tleads_mtr[j1, j2] += tamp
    return tleads_mtr

def make_tleads_dict(tleads, si):
    """
    Makes single particle tunneling amplitude dictionary.

    Parameters
    ----------
    tleads : list, dict, or array
        Contains single particle tunneling amplitudes.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.

    Returns
    -------
    tleads_dict : dictionary
        Dictionary containing tunneling amplitudes.
        tleads[(lead, state)] gives the tunneling amplitude.
    """
    if isinstance(tleads, dict):
        tleads_dict = tleads
    elif isinstance(tleads, list):
        tleads_dict = {}
        for j0 in tleads:
            j1, j2, tamp = j0
            tleads_dict.update({(j1, j2):tamp})
    elif isinstance(tleads, np.ndarray):
        nleads, nsingle = tleads.shape
        tleads_dict = {}
        for j1 in range(nleads):
            for j2 in range(nsingle):
                if tleads[j1, j2] != 0:
                    tleads_dict.update({(j1, j2):tleads[j1, j2]})
    #
    if si.symmetry is 'spin':
        tleads_dict_spin = dict(tleads_dict)
        for j0 in tleads_dict:
            j1, j2 = j0
            tamp = tleads_dict[j0]
            tleads_dict_spin.update({(j1+si.nleads_sym,
                                      j2+si.nsingle_sym):tamp})
        return tleads_dict_spin
    else:
        return tleads_dict

def make_array(lst_old, lst, si, npar=None):
    """
    Converts dictionary or list of mulst, tlst or dlst to an array.

    Parameters
    ----------
    lst_old,lst : list, dict, or array
        Contains lead parameters.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.

    Returns
    -------
    lst_arr : array
        Numpy array containing lead parameters.
    """
    #
    if npar is None:
        npar = si.nleads_sym
    if isinstance(lst, dict):
        if lst_old is None:
            lst_arr = np.zeros(npar, dtype=doublenp)
        else:
            lst_arr = lst_old[0:npar]
        for j1 in lst:
            lst_arr[j1] = lst[j1]
    elif isinstance(lst, (list, np.ndarray)):
        lst_arr = np.array(lst, dtype=doublenp)
    elif isinstance(lst, numbers.Number):
        lst_arr = lst*np.ones(npar, dtype=doublenp)
    else:
        lst_arr = np.zeros(npar, dtype=doublenp)
    #
    if si.symmetry is 'spin':
        return np.concatenate((lst_arr, lst_arr))
    else:
        return lst_arr

def make_array_dlst(dlst_old, dlst, si, npar=None):
    if npar is None:
        npar = si.nleads_sym
    if dlst_old is None:
        lst_arr = np.zeros((npar,2), dtype=doublenp)
    else:
        lst_arr = dlst_old[0:npar]
    #
    if isinstance(dlst, numbers.Number):
        lst_arr[:,0] = -dlst*np.ones(npar, dtype=doublenp)
        lst_arr[:,1] = +dlst*np.ones(npar, dtype=doublenp)
    elif isinstance(dlst, dict):
        for j1 in dlst:
            if isinstance(dlst[j1], numbers.Number):
                lst_arr[j1] = (-dlst[j1], dlst[j1])
            else:
                lst_arr[j1] = dlst[j1]
    elif isinstance(dlst, (list, np.ndarray)):
        if isinstance(dlst[0], numbers.Number):
            lst_arr[:,0] = -np.array(dlst, dtype=doublenp)
            lst_arr[:,1] = +np.array(dlst, dtype=doublenp)
        else:
            lst_arr = np.array(dlst, dtype=doublenp)
    #
    if si.symmetry is 'spin':
        return np.concatenate((lst_arr, lst_arr))
    else:
        return lst_arr
#---------------------------------------------------------------------------------------------------

class LeadsTunneling(object):
    """
    Class for defining tunneling amplitude from leads to the quantum dot.

    Attributes
    ----------
    nleads : int
        Number of the leads.
    tleads : dict, list or array
        Dictionary, list or numpy array defining single particle tunneling amplitudes.
        numpy array has to be nleads by nsingle.
    si : StateIndexing
        StateIndexing or StateIndexingDM object.
    mulst : dict, list or array
        Dictionary, list or numpy array containing chemical potentials of the leads.
    tlst : dict, list or array
        Dictionary, list or numpy array containing temperatures of the leads.
    dlst : dict, list or array
        Dictionary, list or numpy array containing bandwidths of the leads.
    mtype : type
        Defines type of Tba0 and Tba matrices. For example, float, complex, etc.
    Tba0 : array
        nleads by nmany by nmany array, which contains many-body tunneling amplitude matrix
        in Fock basis.
    Tba : array
        nleads by nmany by nmany array, which contains many-body tunneling amplitude matrix,
        which is used in calculations.
    """

    def __init__(self, nleads, tleads, si, mulst, tlst, dlst, mtype=complex):
        """Initialization of the LeadsTunneling class."""
        si.nleads = nleads
        si.nleads_sym = nleads//2 if si.symmetry is 'spin' else nleads
        #
        self.si = si
        self.tleads = make_tleads_dict(tleads, si)
        self.mulst = make_array(None, mulst, si)
        self.tlst = make_array(None, tlst, si)
        self.dlst = make_array_dlst(None, dlst, si)
        self.mtype = mtype
        self.Tba0 = construct_Tba(self, self.tleads)
        self.Tba = self.Tba0

    def add(self, tleads=None, mulst=None, tlst=None, dlst=None, updateq=True, lstq=True):
        """
        Adds a value to single particle tunneling amplitudes and correspondingly redefines
        many-body tunneling matrix Tba.

        Parameters
        ----------
        tleads,mulst,tlst,dlst : dict
            Dictionaries describing what values to add.
            For example, tleads[(lead, state)] = value to add.
        updateq : bool
            Specifies if the values of the single particle amplitudes will be updated.
            The many-body tunneling amplitudes Tba will be updates in either case.
        lstq : bool
            Determines if the values will be added to mulst, tlst, dlst.
        """
        if lstq:
            if mulst is not None:
                self.mulst += make_array(None, mulst, self.si)
            if tlst is not None:
                self.tlst += make_array(None, tlst, self.si)
            if dlst is not None:
                self.dlst += make_array_dlst(None, dlst, self.si)
        if tleads is not None:
            if updateq:
                tleads = make_tleads_dict(tleads, self.si)
                for j0 in tleads:
                    try:    self.tleads[j0] += tleads[j0]       # if tleads[j0] != 0:
                    except: self.tleads.update({j0:tleads[j0]}) # if tleads[j0] != 0:
            self.Tba0 = construct_Tba(self, tleads, self.Tba0)

    def change(self, tleads=None, mulst=None, tlst=None, dlst=None):
        """
        Changes the values of the single particle tunneling amplitudes and correspondingly redefines
        many-body tunneling matrix Tba.

        Parameters
        ----------
        tleads,mulst,tlst,dlst : dict
            Dictionaries describing what values to change.
            For example, tleads[(lead, state)] = the new value.
        """
        if mulst is not None:
            self.mulst = make_array(self.mulst, mulst, self.si)
        if tlst is not None:
            self.tlst = make_array(self.tlst, tlst, self.si)
        if dlst is not None:
            self.dlst = make_array_dlst(self.dlst, dlst, self.si)
        #
        if tleads is not None:
            tleads = make_tleads_dict(tleads, self.si)
            # Find the differences from the previous tunneling amplitudes
            tleads_add = {}
            for j0 in tleads:
                try:
                    tleads_diff = tleads[j0]-self.tleads[j0]
                    if tleads_diff != 0:
                        tleads_add.update({j0:tleads_diff})
                        self.tleads[j0] += tleads_diff
                except:
                    tleads_diff = tleads[j0]
                    if tleads_diff != 0:
                        tleads_add.update({j0:tleads_diff})
                        self.tleads.update({j0:tleads_diff})
            # Add the differences
            self.add(tleads_add, updateq=False, lstq=False)

    def rotate(self, vecslst, indexing='n'):
        """
        Rotates tunneling amplitude matrix Tba0 in Fock basis to Tba,
        which is in eigenstate basis of the quantum dot.

        Parameters
        ----------
        veclst : list of arrays
            List of size ncharge containing arrays defining eigenvector matrices for given charge.
        indexing : string
            Specifies what kind of rotation procedure to use. Default is si.indexing.
        """
        self.Tba = rotate_Tba(self.Tba0, vecslst, self.si, indexing, self.mtype)

    def use_Tba0(self):
        """
        Sets the Tba matrix for calculation to Tba0 in the Fock basis.
        """
        self.Tba = self.Tba0

    def update_Tba0(self, nleads, tleads, mtype=complex):
        """
        Updates the Tba0 in the Fock basis using new single-particle tunneling amplitudes.

        Parameters
        ----------
        tleads : array
            The new single-particle tunneling amplitudes. See attribute tleads.
        """
        si = self.si
        si.nleads = nleads
        si.nleads_sym = nleads//2 if si.symmetry is 'spin' else nleads
        self.tleads = make_tleads_dict(tleads, si)
        self.mtype = mtype
        self.Tba0 = construct_Tba(self, tleads)
